-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local string = _G.string
local math = _G.math
local table = _G.table

local tonumber = _G.tonumber
local tostring = _G.tostring
local time = _G.time
local type = _G.type

local wipe = table.wipe
local tinsert = table.insert
local tremove = table.remove

local ipairs, pairs = _G.ipairs, _G.pairs

-------------------------------------------------------------------------------
-- Localized Blizzard API
-------------------------------------------------------------------------------
local UnitClass, UnitLevel, UnitName = _G.UnitClass, _G.UnitLevel, _G.UnitName
local GetItemInfo, GetItemGem = _G.GetItemInfo, _G.GetItemGem
local GetInventoryItemLink = _G.GetInventoryItemLink
local GetQuestDifficultyColor = _G.GetQuestDifficultyColor

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local MODNAME = "GearGauge"

local LibStub = _G.LibStub
local GearGauge = LibStub("AceAddon-3.0"):NewAddon(MODNAME, "AceEvent-3.0", "AceConsole-3.0")

local LDB = LibStub("LibDataBroker-1.1")
local QTip = LibStub("LibQTip-1.0")
local LC = LOCALIZED_CLASS_NAMES_MALE
local LIB = LibStub("LibItemBonus-2.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]
local L = LibStub("AceLocale-3.0"):GetLocale(MODNAME, "enUS", true, dev)

local DataObj

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local MAX_PLAYER_LEVEL = 80	-- Override the global one, since people without WoTLK will have it set to 70 (or 60 without TBC). Sad, I know.
local TBC_MINLEV = 61
local WOTLK_MINLEV = 71
local IS_ENCHANTER = false
local MY_CLASS = select(2, UnitClass("player"))

local COLOR_RED = "|cffff0000"
local COLOR_TABLE = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
local CLASS_COLORS = {}

for k, v in pairs(COLOR_TABLE) do
	CLASS_COLORS[k] = string.format("%2x%2x%2x", v.r * 255, v.g * 255, v.b * 255)
end
GearGauge.player_color = COLOR_TABLE[MY_CLASS]
GearGauge.CLASS_COLORS = CLASS_COLORS

local EQUIPMENT = {
	["Head"]		= { minlev = 30,	name = _G["INVTYPE_HEAD"] },
	["Neck"]		= { minlev = 30,	name = _G["INVTYPE_NECK"] },
	["Shoulder"]		= { minlev = 20,	name = _G["INVTYPE_SHOULDER"] },
	["Shirt"]		= { minlev = 1,		name = _G["INVTYPE_BODY"] },
	["Chest"]		= { minlev = 1,		name = _G["INVTYPE_CHEST"] },
	["Waist"]		= { minlev = 1,		name = _G["INVTYPE_WAIST"] },
	["Legs"]		= { minlev = 1,		name = _G["INVTYPE_LEGS"] },
	["Feet"]		= { minlev = 1,		name = _G["INVTYPE_FEET"] },
	["Wrist"]		= { minlev = 1,		name = _G["INVTYPE_WRIST"] },
	["Hands"]		= { minlev = 1,		name = _G["INVTYPE_HAND"] },
	["Finger0"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"] },
	["Finger1"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"] },
	["Trinket0"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"] },
	["Trinket1"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"] },
	["Back"]		= { minlev = 1,		name = _G["INVTYPE_CLOAK"] },
	["MainHand"]		= { minlev = 1,		name = _G["INVTYPE_WEAPONMAINHAND"] },
	["SecondaryHand"]	= { minlev = 1,		name = _G["INVTYPE_WEAPONOFFHAND"] },
	["Ranged"]		= { minlev = 1,		name = _G["INVTYPE_RANGED"] },
	["Tabard"]		= { minlev = 1,		name = _G["INVTYPE_TABARD"] },
}

local ORDERED_EQUIPMENT = {
	[1]	= "Head",	[2]	= "Neck",		[3]	= "Shoulder",
	[4]	= "Shirt",	[5]	= "Chest",		[6]	= "Waist",
	[7]	= "Legs",	[8]	= "Feet",		[9]	= "Wrist",
	[10]	= "Hands",	[11]	= "Finger0",		[12]	= "Finger1",
	[13]	= "Trinket0",	[14]	= "Trinket1",		[15]	= "Back",
	[16]	= "MainHand",	[17]	= "SecondaryHand",	[18]	= "Ranged",
	[19]	= "Tabard"
}

local TWO_HANDED_WEAPON = {
	[L["Fishing Poles"]] = true,
	[L["Polearms"]] = true,
	[L["Staves"]] = true,
	[L["Two-Handed Axes"]] = true,
	[L["Two-Handed Maces"]] = true,
	[L["Two-Handed Swords"]] = true,
}

local GEM_TYPES = {
	"Blue",
	"Meta",
	"Red",
	"Yellow"
}

local BONUS_VALUE = {
	-------------------------------------------------------------------------------
	-- Enchantments
	-------------------------------------------------------------------------------
	["ARMOR_BONUS"]			= 0.01,		["DEFENSE"]			= 0.25,
	["FISHING"]			= 0,		["PARRY"]			= 1,
	["SPIRIT"]			= 0.2,

	-------------------------------------------------------------------------------
	-- Stats
	-------------------------------------------------------------------------------
	["AGI"]				= 0.2,		["BASE_ARMOR"]			= 0.01,
	["INT"]				= 0.2,		["SPI"]				= 0.2,
	["STA"]				= 0.2,		["STR"]				= 0.2,

	-------------------------------------------------------------------------------
	-- Power
	-------------------------------------------------------------------------------
	["ATTACKPOWER"]			= 1,		["ATTACKPOWERBEAST"]		= 0.1,
	["ATTACKPOWERDEMON"]		= 0.1,		["ATTACKPOWERDRAGON"]		= 0.1,
	["ATTACKPOWERELEMENTAL"]	= 0.1,		["ATTACKPOWERFERAL"]		= 0.014,
	["ATTACKPOWERUNDEAD"]		= 0.1,		["SPELLPOWER"]			= 1,

	-------------------------------------------------------------------------------
	-- Combat ratings.
	-------------------------------------------------------------------------------
	["CR_ARMOR_PENETRATION"]	= 0.1,		["CR_BLOCK"]			= 1,
	["CR_CRIT"]			= 1,		["CR_DEFENSE"]			= 0.25,
	["CR_DODGE"]			= 0.25,		["CR_EXPERTISE"]		= 0.1,
	["CR_HASTE"]			= 1,		["CR_HIT"]			= 1,
	["CR_PARRY"]			= 1,		["CR_RANGEDCRIT"]		= 1,
	["CR_RANGEDHASTE"]		= 1,		["CR_RANGEDHIT"]		= 1,
	["CR_RESILIENCE"]		= 0.1,

	-------------------------------------------------------------------------------
	-- Base values
	-------------------------------------------------------------------------------
	["BLOCKVALUE"]			= 1,		["HEALTH"]			= 0.1,
	["MANA"]			= 0.1,

	-------------------------------------------------------------------------------
	-- Regeneration
	-------------------------------------------------------------------------------
	["HEALTHREG"]			= 0.35,		["MANAREG"]			= 0.35,

	-------------------------------------------------------------------------------
	-- Resistances
	-------------------------------------------------------------------------------
	["ARCANERES"]			= 0.125,	["FIRERES"]			= 0.125,
	["FROSTRES"]			= 0.125,	["HOLYRES"]			= 0.125,
	["NATURERES"]			= 0.125,	["SHADOWRES"]			= 0.125,

	-------------------------------------------------------------------------------
	-- Percentage modifiers
	-------------------------------------------------------------------------------
	["MOD_ARMOR"]			= 1,		["MOD_DAMAGE_CRIT"]		= 1,
	["MOD_HEAL_CRIT"]		= 1,		["MOD_INT"]			= 1,
	["MOD_MANA"]			= 1,		["MOD_SPELLDAMAGE_TAKEN"]	= 1,
	["MOD_STA"]			= 1,

	-------------------------------------------------------------------------------
	-- Reductions
	-------------------------------------------------------------------------------
	["FEARREDUCTION"]		= 0.1,		["DISARMREDUCTION"]		= 0.1,
	["SILENCEREDUCTION"]		= 0.1,		["SNAREREDUCTION"]		= 0.1,
	["STUNREDUCTION"]		= 0.1,		["THREATREDUCTION"]		= 1,

	-------------------------------------------------------------------------------
	-- Miscellaneous
	-------------------------------------------------------------------------------
	["RUNSPEED"]			= 0.1,		["SPELLPEN"]			= 0.1,
}

local CLASS_NUMBER = {
	["DEATHKNIGHT"]	= 1,
	["DRUID"]	= 2,
	["HUNTER"]	= 3,
	["MAGE"]	= 4,
	["PALADIN"]	= 5,
	["PRIEST"]	= 6,
	["ROGUE"]	= 7,
	["SHAMAN"]	= 8,
	["WARLOCK"]	= 9,
	["WARRIOR"]	= 10,
}

GearGauge.ORDERED_CLASSES = {
	[1]	= "DEATHKNIGHT",
	[2]	= "DRUID",
	[3]	= "HUNTER",
	[4]	= "MAGE",
	[5]	= "PALADIN",
	[6]	= "PRIEST",
	[7]	= "ROGUE",
	[8]	= "SHAMAN",
	[9]	= "WARLOCK",
	[10]	= "WARRIOR",
}
local ORDERED_CLASSES = GearGauge.ORDERED_CLASSES

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local player_equip		-- Player's equipment list.
local db

-------------------------------------------------------------------------------
-- Table cache
-------------------------------------------------------------------------------
--[===[@debug@
local used_tables = 0
local cached_tables = 0
local table_locs = {}
--@end-debug@]===]

local AcquireTable, ReleaseTable
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable(where)
		local tbl = tremove(table_cache) or {}
		--[===[@debug@
		used_tables = used_tables + 1
		cached_tables = #table_cache
		table_locs[tbl] = where
		--@end-debug@]===]
		return tbl
	end

	-- Cleans the table and stores it in the cache.
	function ReleaseTable(tbl)
		if not tbl then return false end

		wipe(tbl)
		tinsert(table_cache, tbl)
		--[===[@debug@
		used_tables = used_tables - 1
		cached_tables = #table_cache
		table_locs[tbl] = nil
		--@end-debug@]===]
		return true
	end
end	-- do

-------------------------------------------------------------------------------
-- Local functions
-------------------------------------------------------------------------------
function GearGauge:debug(msg)
	--[===[@debug@
	print(string.format("%sDebug|r: %s", COLOR_RED, tostring(msg)))
	--@end-debug@]===]
end

-- Finds an entry in the entries table by name.
local function FindEntry(entry)
	local found_name, search_name
	entry = string.lower(entry)

	for name in pairs(db.realm.entries) do
		search_name = string.lower(name)

		if search_name == entry then
			found_name = name
			break
		end
	end
	return found_name
end

-- Releases all of the sub-tables in the equipment list, then releases the list.
local function ReleaseItems(item_list)
	if not item_list then
		GearGauge:debug("No item list.")
		return
	end

	local count = 0
	for idx, slot in ipairs(ORDERED_EQUIPMENT) do
		if ReleaseTable(item_list[slot]) then
			count = count + 1
		end
	end
	if ReleaseTable(item_list) then
		count = count + 1
	end
end

local function AverageItemLevels(equip_list)
	local level_cnt = 0
	local item_cnt = 0

	for slot in pairs(EQUIPMENT) do
		local item = equip_list[slot]

		if item.level > 0 and slot ~= "Shirt" and slot ~= "Tabard" then
			level_cnt = level_cnt + item.level
			item_cnt = item_cnt + 1
		end
	end
	return (item_cnt > 0) and math.floor(level_cnt / item_cnt) or 0
end

local function CanScan(unit, silent)
	if unit == "player" or UnitName(unit) == UnitName("player") then return false end

	local prefs = db.global.scanning

	if UnitFactionGroup(unit) ~= UnitFactionGroup("player") and not prefs.enemies then return false end

	if (InCombatLockdown() or not CheckInteractDistance(unit, 1)
	    or not (UnitExists(unit) and UnitIsPlayer(unit) and (UnitLevel(unit) >= prefs.minlevel)
		    and (UnitLevel(unit) <= prefs.maxlevel))) then return false end

	if not CanInspect(unit, false) then
		if not silent then
			GearGauge:Print(string.format("%s%s: %s|r", COLOR_RED, UnitName(unit), ERR_PLAYER_WRONG_FACTION))
		end
		return false
	end
	return true
end

-------------------------------------------------------------------------------
-- Item status generation
-------------------------------------------------------------------------------
local BuildGearData
do
	-------------------------------------------------------------------------------
	-- Check to see if an empty slot is OK given the unit's level/class.
	-------------------------------------------------------------------------------
	local function IsEmptySlotOK(slot, level, class)
		if slot == "Shirt" or slot == "Tabard" then return true end

		if slot == "SecondaryHand"  then
			if level < 5 and (class == LC["PALADIN"] or class == LC["SHAMAN"] or class == LC["WARRIOR"]) then	-- Shields
				return true
			else
				return false
			end
		end

		if slot == "Ranged" then
			if  level < 52 and (class == LC["DRUID"] or class == LC["PALADIN"] or class == LC["SHAMAN"]) then	-- Idols, Librams, Totems
				return true
			elseif level < 5 and (class == LC["MAGE"] or class == LC["PRIEST"] or class == LC["WARLOCK"]) then	-- Wands
				return true
			else
				return false
			end
		end
		return EQUIPMENT[slot].minlev > level
	end

	-- Item subtypes which are not enchantable
	local NON_ENCHANTABLE = {
		[L["Idols"]] = true,
		[L["Librams"]] = true,
		[L["Miscellaneous"]] = true,
		[L["Other"]] = true,
		[L["Sigils"]] = true,
		[L["Thrown"]] = true,
		[L["Totems"]] = true,
		[L["Wands"]] = true,
	}
	local prev_stype	-- Placeholder in case previous slot contained a two-handed weapon.
	local current_type	-- Type of "unit" passed to BuildGearData() and ConstructItem().

	-------------------------------------------------------------------------------
	-- Assign values to an item for later use.
	-------------------------------------------------------------------------------
	local function ConstructItem(item_link, slot, unit)
		local unit_level, unit_class
		local item = AcquireTable("ConstructItem")

		if current_type ~= "table" then
			unit_level = UnitLevel(unit)
			unit_class = UnitClass(unit)
		else
			unit_level = unit.level
			unit_class = unit.class
		end
		-- Default values here
		item.slot, item.rating, item.level = slot, 0, 0
		item.emptyMeta, item.emptyRed, item.emptyYellow, item.emptyBlue = 0, 0, 0, 0
		item.ench_id = -1

		if not item_link then
			local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]
			local is_ok = IsEmptySlotOK(slot, unit_level, unit_class)

			if not two_handed and not is_ok then
				item.rating = 0 - (unit_level * 0.5)
			end
			item.empty_ok = is_ok
			prev_stype = nil
			return item
		end
		item.name, item.link, item.rarity, item.level, item.minlev, item.type, item.stype, item.stacksize, item.equiploc, item.texture = GetItemInfo(item_link)
		prev_stype = item.stype

		-- Set the level of Heirloom items depending on the player's level
		if item.rarity == 7 then
			if unit_level <= TBC_MINLEV then
				item.level = unit_level
			elseif unit_level > TBC_MINLEV and unit_level <= WOTLK_MINLEV then
				item.level = 3 * (unit_level - 64) + 90
			elseif unit_level > WOTLK_MINLEV then
				item.level = 4 * (unit_level - 73) + 142
			end
		end
		if not item.level then item.level = 0 end	-- Shouldn't happen, but it did.
		-- Must check for Tabard and Shirt here since they may have green text which LibItemBonus cannot properly parse, yet will still allocate a bonus table for.
		if item.level <= 1 or slot == "Tabard" or slot == "Shirt" then return item end

		item.info = LIB:ScanItemLink(item_link)

		local bonuses = item.info.bonuses
		local points = 0

                if bonuses then
			-- Get gem information
			item.emptyBlue = bonuses["EMPTY_SOCKET_BLUE"] or 0
			bonuses["EMPTY_SOCKET_BLUE"] = nil

			item.emptyMeta = bonuses["EMPTY_SOCKET_META"] or 0
			bonuses["EMPTY_SOCKET_META"] = nil

			item.emptyRed = bonuses["EMPTY_SOCKET_RED"] or 0
			bonuses["EMPTY_SOCKET_RED"] = nil

			item.emptyYellow = bonuses["EMPTY_SOCKET_YELLOW"] or 0
			bonuses["EMPTY_SOCKET_YELLOW"] = nil

			-- Remove these - they are duplicates.
			bonuses["CR_SPELLHASTE"] = nil
			bonuses["CR_SPELLCRIT"] = nil
			bonuses["CR_SPELLHIT"] = nil

			-- Remove these
			bonuses["WEAPON_SPEED"] = nil
			bonuses["WEAPON_MIN"] = nil
			bonuses["WEAPON_MAX"] = nil

			local count = 0
                        for k, v in pairs(bonuses) do
				local bonus = LIB:GetBonus(k, bonuses, unit_level)

				if BONUS_VALUE[k] then bonus = bonus * BONUS_VALUE[k] end

				points = points + bonus
				count = count + 1
                        end

			if count == 0 then	-- The item has bonuses, but LibItemBonus-2.0 couldn't parse them, Give benefit of the doubt.
				points = points + (item.level * 0.5)
			end
                end
		points = ((points + (item.level / unit_level)) * 0.1)

		local rating = points
		local _, _, ench_id = string.split(":", item_link)	-- Grab the item's enchant id, if enchanted.

		if slot ~= "Waist" and slot ~= "Shirt" and slot ~= "Tabard" then
			item.ench_id = ench_id

			if ench_id == nil or ench_id == "0" then
				if ((slot == "Finger0" or slot == "Finger1") and (unit == "player" and IS_ENCHANTER)) or not NON_ENCHANTABLE[item.stype] then
					rating = rating - (points * 0.25)
				else
					item.ench_id = -1
				end
			end
		end

		for k, v in ipairs(GEM_TYPES) do
			if item["empty"..v] > 0 then
				rating = rating - ((points * 0.125) * item["empty"..v])
			end
		end
		item.rating = rating

		return item
	end

	-------------------------------------------------------------------------------
	-- List of items equipped by unit.
	-------------------------------------------------------------------------------
	function BuildGearData(unit, item_list, link_list)
		local rating = 0

		current_type = type(unit)	-- Set the current unit type for ConstructItem().
		prev_stype = nil	-- Nil this out - it's used in ConstructItem().

		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			local link

			if current_type ~= "table" then
				link = GetInventoryItemLink(unit, EQUIPMENT[slot].id)
			else
				link = unit.link_list[slot]
			end
			local item = ConstructItem(link, slot, unit)

			rating = rating + item.rating
			item_list[slot] = item

			if link_list then
				link_list[slot] = link
			end
		end
		return tonumber(string.format("%.2f", rating))
	end
end -- do

do
	local function CreatePlayerInfo(unit)
		local info = AcquireTable("CreatePlayerInfo")
		local local_name, class_name = UnitClass(unit)

		db.global.localized_classes[class_name] = local_name

		info.class = CLASS_NUMBER[class_name]
		info.faction = UnitFactionGroup(unit)
		info.guild = GetGuildInfo(unit)
		info.link_list = AcquireTable("CreatePlayerInfo (link_list)")
		info.level = UnitLevel(unit)
		info.race = UnitRace(unit)
		return info
	end

	function GearGauge:Scan(unit)
		if not CanScan(unit, false) then
			return nil, nil
		end
		NotifyInspect(unit)

		local prefs = db.global.scanning
		local unit_name = UnitName(unit)
		local target = db.realm.entries[unit_name]

		if not target then
			target = CreatePlayerInfo(unit)
			target.scanned = time()
			db.realm.entries[unit_name] = target
			self.panel:Refresh(true)
		else
			local scanned = time() - target.scanned

			-- Update anything that may have changed since last scan attempt.
			target.level = UnitLevel(unit)
			target.guild = GetGuildInfo(unit)

			-- If the delay hasn't been reached yet, don't rescan.
			if scanned < (60 * prefs.delay) then
				return target.rating, target.equip_level
			else
				target.scanned = time()
			end
		end
		local item_list = AcquireTable("GearGauge:Scan")
		local rating = BuildGearData(unit, item_list, target.link_list)
		local equip_level = AverageItemLevels(item_list)

		target.rating = rating
		target.equip_level = equip_level

		ReleaseItems(item_list)

		if not prefs.notify then
			return rating, equip_level
		end
		local color_str = ""

		if UnitFactionGroup(unit) ~= UnitFactionGroup("player") then
			color_str = COLOR_RED
		end

		self:Print(string.format("%s%s %s: %.2f. %s %s: %d|r", color_str, unit_name, RATING, rating, ITEMS, LEVEL, equip_level))
		return rating, equip_level
	end
end -- do

function GearGauge:RemoveEntry(entry, notify)
	ReleaseTable(db.realm.entries[entry].link_list)
	ReleaseTable(db.realm.entries[entry])
	db.realm.entries[entry] = nil

	if notify then
		self:Print(string.format(L["ENTRY_DELETE_MSG"], entry))
	end
end

-- Rebuild the player's gear rating, and update the DataObj text.
local function UpdateDataObj()
	ReleaseTable(player_equip)
	player_equip = AcquireTable("UpdateDataObj")

	DataObj:SetIcon()
	DataObj.text = string.format("%.2f", BuildGearData("player", player_equip, nil))
end

-------------------------------------------------------------------------------
-- Tooltip functions
-------------------------------------------------------------------------------
local tipdata = {}	-- So the tooltip can be refreshed elsewhere.

local ShowTooltip
do
	-- Returns a string based on whether or not the item is enchanted
	local function GetItemStatus(item)
		if not item.link then return HELPFRAME_ITEM_TITLE end

		local ench_id = item.ench_id
		local retstr = ""

		if ench_id ~= -1 and (ench_id == nil or ench_id == "0") then
			retstr = retstr.."|TInterface\\Icons\\Trade_Engraving:22:22|t"
		end
		local socket_tex = "|TInterface\\ItemSocketingFrame\\UI-EmptySocket-"

		for k, v in ipairs(GEM_TYPES) do
			if item["empty"..v] > 0 then
				retstr = retstr..socket_tex..v..":22:22|tx"..item["empty"..v]
			end
		end
		return retstr
	end
	local bonus_tip

	local function ItemOnEnter(cell, arg)
		if not arg then return end

		bonus_tip = QTip:Acquire(MODNAME.."BonusTip", 2, "LEFT", "LEFT")
		bonus_tip:ClearAllPoints()
		bonus_tip:SetPoint("TOPRIGHT", GearGauge.tooltip, "TOPLEFT")
		bonus_tip:Clear()

		local line, column = bonus_tip:AddLine()
		local count = 0
		bonus_tip:SetCell(line, column, L["Item Bonuses"], "CENTER", 2)
		bonus_tip:AddSeparator()

		for k, v in pairs(arg) do
			bonus_tip:AddLine(BONUS_VALUE[k] and L[k] or k, v)
			count = count + 1
		end

		if count == 0 then
			line, column = bonus_tip:AddLine()
			bonus_tip:SetCell(line, column, UNKNOWN, "CENTER", 2)
		end
		if TipTac and TipTac.AddModifiedTip then TipTac:AddModifiedTip(bonus_tip, true) end

		bonus_tip:Show()
		bonus_tip:SetScale(db.global.tooltip.scale)
		GearGauge.tooltip:SetScale(db.global.tooltip.scale)
	end

	local function ItemOnLeave(cell, arg)
		bonus_tip = QTip:Release(bonus_tip)
	end

	function ShowTooltip(anchor, unit, center, refresh)
		local prefs = db.global.tooltip

		GearGauge.tooltip = QTip:Acquire(MODNAME.."Tooltip", 4, "LEFT", "LEFT", "LEFT")

		local tooltip = GearGauge.tooltip

		if not refresh then
			tooltip:ClearAllPoints()

			if not center then
				tooltip:SetAutoHideDelay(prefs.timer, anchor)
				tooltip:SmartAnchorTo(anchor)
			else
				tooltip:SetPoint("CENTER", anchor, 0, 0)
			end
		end
		tooltip:Clear()
		tooltip:EnableMouse(true)

		if TipTac and TipTac.AddModifiedTip then
			TipTac:AddModifiedTip(tooltip, true)
		end

		local item_list
		local unit_name, unit_class, unit_race, unit_level, unit_guild

		unit = unit or "player"

		-- This is set so that the tooltip can be refreshed elsewhere.
		tipdata.anchor = anchor
		tipdata.unit = unit
		tipdata.center = center

		if unit == "player" then
			item_list = player_equip
			unit_name = UnitName(unit)
			unit_class = CLASS_NUMBER[MY_CLASS]
			unit_race = UnitRace(unit)
			unit_level = UnitLevel(unit)
			unit_guild = GetGuildInfo(unit)
		else
			item_list = AcquireTable("ShowTooltip")

			if unit == "target" then
				unit_name = UnitName("target")
			else
				unit_name = unit	-- Not player or target, so name was passed as unit.
			end

			local entry = db.realm.entries[unit_name]
			BuildGearData(entry, item_list)
			unit_level = entry.level
			unit_class = entry.class
			unit_race = entry.race
			unit_guild = entry.guild
		end
		local rating_tot = 0
		local diff = GetQuestDifficultyColor(unit_level)
		local class_english = ORDERED_CLASSES[unit_class]
		local class_name = db.global.localized_classes[class_english] or class_english

		tooltip:SetCellMarginV(1)
		tooltip:SetCellMarginH(0)

		local line, column = tooltip:AddHeader()

		tooltip:SetCell(line, column,
				string.format("%s - %s |cff%2x%2x%2x%d|r %s |cff%s%s|r", unit_name, LEVEL, diff.r * 255, diff.g * 255, diff.b * 255, unit_level, unit_race,
					      CLASS_COLORS[class_english], class_name),
				"CENTER", 4)

		if unit_guild then
			line, column = tooltip:AddHeader()
			tooltip:SetCell(line, column, string.format("<%s>", unit_guild), "CENTER", 4)
		end
		tooltip:AddSeparator()
		line, column = tooltip:AddHeader("", "", RATING, ADDON_MISSING)

		tooltip:SetCell(line, 1, ITEMS, "CENTER", 2)
		tooltip:AddSeparator()

		local two_handed = nil
		local prev_stype = nil

		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			local item = item_list[slot]

			rating_tot = rating_tot + item.rating

			if item.link then
				local r, g, b, hex = GetItemQualityColor(item.rarity)

				line, column = tooltip:AddLine()

				if prefs.coloring == "Line" or prefs.coloring == "Both" then
					tooltip:SetLineColor(line, r, g, b)
				end
				local item_name = item.name

				if prefs.coloring == "Text" or prefs.coloring == "Both" then
					item_name = hex..item_name.."|r"
				end
				tooltip:SetCell(line, 1, "|T"..item.texture..":22:22|t")
				tooltip:SetCell(line, 2, item_name)

				if item.info and item.info.bonuses then
					tooltip:SetLineScript(line, "OnEnter", ItemOnEnter, item.info.bonuses)
					tooltip:SetLineScript(line, "OnLeave", ItemOnLeave, nil)
				end
				tooltip:SetCell(line, 3, string.format("%.2f", item.rating))
				tooltip:SetCell(line, 4, GetItemStatus(item))
				prev_stype = item.stype
			else
				local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]

				if not two_handed  then
					line, column = tooltip:AddLine()

					if not item.empty_ok then
						tooltip:SetLineColor(line, 1)
					end
					tooltip:SetCell(line, 1, EQUIPMENT[slot].name, nil, "LEFT", 2)
					tooltip:SetCell(line, 3, string.format("%.2f", item.rating))
					tooltip:SetCell(line, 4, GetItemStatus(item))
					prev_stype = nil
				end
			end
		end
		tooltip:AddSeparator()

		line, column = tooltip:AddLine()

		tooltip:SetCell(line, 1, string.format("%s: %d", LEVEL, AverageItemLevels(item_list)), "LEFT", 2)
		tooltip:SetCell(line, 3, string.format("%.2f", rating_tot))
		tooltip:Show()
		tooltip:SetScale(db.global.tooltip.scale)

		if unit ~= "player" then
			ReleaseItems(item_list)
		end
	end
end	-- do

-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------
local OpenConfig
local DB_VERSION = 0.1

function GearGauge:OnInitialize()
	-- Set the numeric ID and texture strings for the equipment slots
	for slot in pairs(EQUIPMENT) do
		EQUIPMENT[slot].id, EQUIPMENT[slot].texture = GetInventorySlotInfo(slot.."Slot")
	end
	DataObj = LDB:NewDataObject(MODNAME, {
		type	= "data source",
		label	= MODNAME,
		text	= "0",
		SetIcon = function(self)
				  local num = math.random(8)
				  self.icon = "Interface\\Icons\\INV_Misc_Gear_0"..num
			  end,
		OnEnter = function(display, motion) ShowTooltip(display) end,
		OnLeave = function(display, motion) end,
		OnClick = function(display, button)
				  if button == "LeftButton" then
					  GearGauge.panel:Open(nil, true)
				  else
					  OpenConfig()
				  end
			  end,
	})
	DataObj:SetIcon()

	local defaults = {
		global = {
			minimap_icon		= {
				hide		= true,
			},
			scanning = {
				delay		= 10,
				persistence	= 30,
				maxlevel	= MAX_PLAYER_LEVEL,
				minlevel	= 1,
				notify		= true,
				enemies		= true,
			},
			tooltip = {
				coloring	= "Both",
				timer		= 0.25,
				scale		= 1.0,
			},
			-- This is filled in as we go. Yes, it is hackish.
			localized_classes = {},
		},
		realm = {
			db_version = 0,
			entries = {},		-- Scanned players.
		},
	}
	db = LibStub("AceDB-3.0"):New(MODNAME.."DB", defaults)
	self.db = db

	LibStub("LibDBIcon-1.0"):Register(MODNAME, DataObj, db.global.minimap_icon)
	self:SetupOptions()
	self:RegisterChatCommand("geargauge", "SlashCommand")
	self:RegisterChatCommand("gg", "SlashCommand")
	self.OnInitialize = nil
end

function GearGauge:OnEnable()
	local enchanting_sklname = GetSpellInfo(7411)	-- Determine if the toon is an enchanter

	for idx = 1, GetNumSkillLines() do
		local skill = GetSkillLineInfo(idx)
		if skill == enchanting_sklname then
			IS_ENCHANTER = true
			break
		end
	end

	if db.realm.db_version < DB_VERSION then
		db.realm.db_version = DB_VERSION
		db.realm.entries = {}
		self:Print("Database change - all entries were removed.")
	end
	UpdateDataObj()
	self:InitializeList()
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
end

function GearGauge:OnDisable()
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("UNIT_INVENTORY_CHANGED")
end

-------------------------------------------------------------------------------
-- Event handlers
-------------------------------------------------------------------------------
function GearGauge:PLAYER_LEVEL_UP()
	UpdateDataObj()
end

function GearGauge:PLAYER_TARGET_CHANGED()
	self:Scan("target")
end

do
	local last_update = 0
	local updater = CreateFrame("Frame", nil, UIParent)

	updater:Hide()
	updater:SetScript("OnUpdate",
			  function(self, elapsed)
				  last_update = last_update + elapsed
				  if last_update >= 0.25 then
					  UpdateDataObj()
					  if GearGauge.tooltip and GearGauge.tooltip:IsVisible() then
						  ShowTooltip(tipdata.anchor, tipdata.unit, tipdata.center, true)
					  end
					  self:Hide()
				  end
			  end)

	function GearGauge:UNIT_INVENTORY_CHANGED(event, unit)
		if unit ~= "player" then return end
		if not updater:IsVisible() then
			last_update = 0
			updater:Show()
		end
	end
end	-- do

-------------------------------------------------------------------------------
-- Subcommands
-------------------------------------------------------------------------------
function GearGauge:ShowPlayerGear(arg1, anchor)
	-- Default to target if no argument was specified.
	if not arg1 or arg1 == "" then
		if not UnitExists("target") then
			GearGauge:Print(ERR_GENERIC_NO_TARGET)
			return
		end
		if not CanScan("target", false) then
			GearGauge:Print(ERR_CLIENT_LOCKED_OUT)
			return
		end
		ShowTooltip(UIParent, "target", true)
		return
	end
	local entry = FindEntry(arg1)

	if not entry then
		GearGauge:Print(string.format(ERR_BAD_PLAYER_NAME_S, arg1))
		return
	end
	ShowTooltip(anchor or UIParent, entry, true)
end

local function RateEntry(entry)
	local name = FindEntry(entry)

	if not name then
		GearGauge:Print(string.format(ERR_BAD_PLAYER_NAME_S, entry))
		return
	end
	local info = db.realm.entries[name]
	local items = AcquireTable("RateEntry")
	local rating = BuildGearData(info, items)
	local equip_level = AverageItemLevels(items)

	local enemies = info.faction ~= UnitFactionGroup("player")
	local color_str = ""
	if enemies then color_str = COLOR_RED end

	GearGauge:Print(color_str..name.." "..RATING..": "..string.format("%.2f", rating)..". "..ITEMS.." "..LEVEL..": "..equip_level.."|r")
	ReleaseItems(items)
end

--[===[@debug@
local ShowTable
do
	local table_lookup = {}

	function ShowTables(arg)
		if used_tables then
			GearGauge:Print(string.format("Used tables: %d", used_tables))
			wipe(table_lookup)

			for k, v in pairs(table_locs) do
				if table_lookup[v] then
					table_lookup[v] = table_lookup[v] + 1
				else
					table_lookup[v] = 1
				end
			end

			for k, v in pairs(table_lookup) do
				GearGauge:debug(string.format("%s: %d", k, v))
			end
		end

		if cached_tables then
			GearGauge:Print(string.format("Cached tables: %d", cached_tables))
		end
	end
end	-- do
--@end-debug@]===]

function OpenConfig()
	local options_frame = InterfaceOptionsFrame

	if options_frame:IsVisible() then
		options_frame:Hide()
	else
		InterfaceOptionsFrame_OpenToCategory(GearGauge.optionsFrame)
	end
end

-------------------------------------------------------------------------------
-- Command table and parser.
-------------------------------------------------------------------------------
do
	local function OpenPanel()
		GearGauge.panel:Open(nil, true)
	end

	local COMMANDS = {
		["config"]	= OpenConfig,
		["panel"]	= OpenPanel,
		["rate"]	= RateEntry,
		["show"]	= GearGauge.ShowPlayerGear,
		--[===[@debug@
		["tables"]	= ShowTables,
		--@end-debug@]===]
	}

	function GearGauge:ShowCommands()
		local str = ""
		for k, v in pairs(COMMANDS) do
			str = string.format("%s\n  %s", str, k)
		end
		self:Print(str)
	end
	-------------------------------------------------------------------------------
	-- Separates input into two parts, by assigning the first word to arg1 and
	-- the rest of the string to arg2
	-------------------------------------------------------------------------------
	local function StrSplit(input)
		if not input then return nil, nil end
		local arg1, arg2, var1

		arg1, var1 = input:match("^([^%s]+)%s*(.*)$")
		arg1 = (arg1 and arg1:lower() or input:lower())

		if var1 then
			local var2
			arg2, var2 = var1:match("^([^%s]+)%s*(.*)$")
			arg2 = (arg2 and arg2:lower() or var1:lower())
		end
		return arg1, arg2
	end

	function GearGauge:SlashCommand(input)
		local cmd, arg1, arg2, arg3
		local str = input

		cmd, str = StrSplit(str)
		arg1, str = StrSplit(str)
		arg2, arg3 = StrSplit(str)

		if not cmd or (cmd and cmd:trim() == "") then
			self:ShowCommands()
			return
		end

		if not COMMANDS[cmd] then
			self:Print(string.format(L["INVALID_CMD_STR"], cmd))
			self:ShowCommands()
		else
			COMMANDS[cmd](arg1, arg2, arg3)
		end
	end
end	-- do
